"""Internet-related functions."""
from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from getpass import getuser
from html import escape
from http import HTTPStatus
from itertools import chain
from os import scandir
from pathlib import Path
from typing import TYPE_CHECKING, Literal, NotRequired, TypedDict, cast
import contextlib
import logging
import plistlib
import re
import urllib
import urllib.parse

from bs4 import BeautifulSoup as Soup, Tag
import keyring
import requests

from .chromium import generate_chrome_user_agent
from .string import hexstr2bytes
from .system import IS_LINUX

if TYPE_CHECKING:
    from .typing import FileDescriptorOrPath, StrPath

__all__ = ('BookmarksDataset', 'BookmarksHTMLAnchorAttributes', 'BookmarksHTMLFolder',
           'BookmarksHTMLFolderAttributes', 'BookmarksHTMLLink', 'RecurseBookmarksHTMLCallback',
           'check_bookmarks_html_urls', 'generate_html_dir_tree', 'parse_bookmarks_html',
           'recurse_bookmarks_html', 'upload_to_imgbb', 'where_from')

log = logging.getLogger(__name__)
KEY_ORIGIN_URL = 'user.xdg.origin.url'
KEY_WHERE_FROMS = 'com.apple.metadata:kMDItemWhereFroms'


def _getxattr(file: FileDescriptorOrPath,
              name: str,
              *,
              follow_symlinks: bool = False) -> bytes:  # pragma: no cover
    from os import getxattr  # noqa: PLC0415
    return getxattr(file, name, follow_symlinks=follow_symlinks)


def where_from(file: FileDescriptorOrPath, *, webpage: bool = False) -> str | None:
    """Determine where a file came from based on metadata in extended attributes."""
    index = 1 if webpage else 0
    attr_value = _getxattr(file, KEY_ORIGIN_URL if IS_LINUX else KEY_WHERE_FROMS).decode()
    if not IS_LINUX:
        return cast('Sequence[str]', plistlib.loads(hexstr2bytes(attr_value)))[index]
    return attr_value


def generate_html_dir_tree(start_dir: StrPath,
                           *,
                           follow_symlinks: bool = False,
                           depth: int = 2) -> str:
    """Generate a HTML directory listing."""
    def recurse_cwd(path: Path, _cur_depth: int = 0) -> Iterator[str]:
        for entry in sorted(sorted(scandir(path), key=lambda x: x.name),
                            key=lambda x: not x.is_dir(follow_symlinks=follow_symlinks)):
            if (entry.is_dir(follow_symlinks=follow_symlinks) and _cur_depth < depth):
                yield ('<li class="dir mui--text-dark mui--text-body2"><details><summary>'
                       f'<code>{escape(entry.name)}/</code></summary><ul>')
                yield from recurse_cwd(Path(entry), _cur_depth=_cur_depth + 1)
                yield '</ul></details></li>'
            else:
                isd = entry.is_dir(follow_symlinks=follow_symlinks)
                class_ = 'file' if not isd else 'dir'
                slash = '' if not isd else '/'
                yield (
                    f'<li class="{class_} mui--text-dark mui--text-body1"><a class="mui--text-dark"'
                    f' href="./{entry.path}"><code>{escape(entry.name)}{slash}</code></a></li>')

    start_dir = Path(start_dir).resolve(strict=True)
    with contextlib.chdir(start_dir):
        title = start_dir.name
        files = ''.join(chain.from_iterable(recurse_cwd(start_dir)))
        return f"""<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="utf-8">
<meta http-equiv="X-UA-Compatible" content="IE=edge">
<meta name="viewport" content="width=device-width, initial-scale=1">
<title>Contents of {title}/</title>
<link rel="stylesheet" href="https://cdn.muicss.com/mui-0.10.3/css/mui.min.css"
crossorigin="anonymous">
<style>
ul {{
    list-style: none;
    padding-inline-start: 1em;
}}
.dir {{
    cursor: pointer;
}}
.mui-appbar {{
    margin-bottom: 2em;
}}
</style>
</head>
<body>
<header class="mui-appbar mui--z1">
<div class="mui-container-fluid">
<h1 class="mui--text-title">Contents of {title}</h1>
</div>
</header>
<div class="mui-container-fluid">
<ul>{files}</ul>
</div>
</body>
</html>"""


def upload_to_imgbb(path: StrPath,
                    *,
                    api_key: str | None = None,
                    keyring_username: str | None = None,
                    timeout: float = 5) -> requests.Response:
    """
    Upload an image to ImgBB.

    Get an API key at https://api.imgbb.com/ and set it with ``keyring set imgbb keyring_username``.
    """
    r = requests.post(
        'https://api.imgbb.com/1/upload',
        files={'image': Path(path).resolve(strict=True).read_bytes()},
        params={'key': api_key or keyring.get_password('imgbb', keyring_username or getuser())},
        timeout=timeout)
    r.raise_for_status()
    return r


def stripped_strings_fixed(child: Tag) -> str:
    return re.sub(r'\s+', ' ', ' '.join(child.stripped_strings))


class BookmarksHTMLAnchorAttributes(TypedDict):
    """Attributes of a bookmark link."""
    add_date: str
    href: str
    icon: NotRequired[str]


class BookmarksHTMLFolderAttributes(TypedDict):
    """Attributes of a bookmark folder."""
    add_date: str
    last_modified: str
    personal_toolbar_folder: NotRequired[Literal['true']]


class BookmarksHTMLLink(TypedDict):
    """A bookmark link."""
    attrs: BookmarksHTMLAnchorAttributes
    title: str
    type: Literal['link']


class BookmarksHTMLFolder(TypedDict):
    """A bookmark folder."""
    attrs: BookmarksHTMLFolderAttributes
    children: list[BookmarksHTMLLink | BookmarksHTMLFolder]
    name: str
    type: Literal['folder']


BookmarksDataset = list[BookmarksHTMLFolder | BookmarksHTMLLink]
"""Bookmark top structure."""

RecurseBookmarksHTMLCallback = Callable[
    [BookmarksHTMLAnchorAttributes, str, Sequence[tuple[str, BookmarksHTMLFolderAttributes]]], None]
"""Callback for :py:func:`recurse_bookmarks_html` to call for each link."""


def recurse_bookmarks_html(soup: Tag, callback: RecurseBookmarksHTMLCallback) -> None:
    """Traverse the ``bookmarks.html`` tree and call the callback for each link."""
    for child in soup.children:
        if not isinstance(child, Tag):
            continue
        match child.name:
            case 'dl' | 'dt' | 'html' | 'body':
                recurse_bookmarks_html(child, callback)
            case 'a':
                folder_path: list[tuple[str, BookmarksHTMLFolderAttributes]] = []
                for parent in child.parents:
                    if parent.name == 'dl' and (h3 := parent.find_previous_sibling('h3')):
                        assert isinstance(h3, Tag)
                        folder_path.append((stripped_strings_fixed(h3),
                                            cast('BookmarksHTMLFolderAttributes', h3.attrs)))
                        break
                callback(cast('BookmarksHTMLAnchorAttributes', child.attrs),
                         stripped_strings_fixed(child), folder_path)


def create_parsed_tree_structure(
        folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]],
        data: BookmarksDataset) -> list[BookmarksHTMLFolder | BookmarksHTMLLink]:
    keys = [f[0] for f in folder_path]
    ref = data
    # This breaks for folders that are named the same at the same depth
    for i, key in enumerate(keys):
        try:
            next(x for x in ref if x['type'] == 'folder' and x['name'] == key)
        except StopIteration:
            new_level: BookmarksHTMLFolder = {
                'attrs': folder_path[i][1],
                'children': [],
                'name': key,
                'type': 'folder'
            }
            ref.append(new_level)
            ref = new_level['children']
    return ref


def parse_bookmarks_html(html_content: str) -> BookmarksDataset:
    """Parse a browser's exported ``bookmarks.html``."""
    from bs4 import BeautifulSoup as Soup  # noqa: PLC0415
    data: BookmarksDataset = []

    def callback(attrs: BookmarksHTMLAnchorAttributes, title: str,
                 folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]]) -> None:
        ref = create_parsed_tree_structure(folder_path, data)
        ref.append({'type': 'link', 'title': title, 'attrs': attrs})

    recurse_bookmarks_html(Soup(html_content, 'html5lib'), callback)
    return data


def check_bookmarks_html_urls(
        html_content: str) -> tuple[BookmarksDataset, BookmarksDataset, BookmarksDataset]:
    """
    Check a browser's exported bookmarks.html's URLs.

    Checks for URLs that are not valid any more (status ``404``) or have changed (statuses ``301``
    and ``302``).
    """
    # After html5lib fixes it, the structure is:
    # DL -> many p -> many DT -> H3 (folder), DL then
    #   p or DT -> A or H3 (folder where structure repeats), DL
    data: BookmarksDataset = []
    changed: BookmarksDataset = []
    not_found: BookmarksDataset = []
    session = requests.Session()
    session.headers.update({
        'cache-control': 'no-cache',
        'dnt': '1',
        'pragma': 'no-cache',
        'referer': 'https://www.google.com/',
        'upgrade-insecure-requests': '1',
        'user_agent': generate_chrome_user_agent()
    })

    def callback(attrs: BookmarksHTMLAnchorAttributes, title: str,
                 folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]]) -> None:
        ref = create_parsed_tree_structure(folder_path, data)
        new_data: BookmarksHTMLLink = {'type': 'link', 'title': title, 'attrs': attrs}
        if 'href' in attrs and re.match(r'^https?://', attrs['href']):
            log.debug('HEAD %s', attrs['href'])
            r = session.head(attrs['href'])
            if r.status_code in {HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND}:
                new_location = r.headers['location']
                if not re.match(r'^https://', new_location):
                    parsed = urllib.parse.urlparse(attrs['href'])
                    port_str = f':{parsed.port}' if parsed.port else ''
                    new_location = f'{parsed.scheme}://{parsed.netloc}{port_str}{new_location}'
                log.info('%d: "%s" @ "%s" -> "%s"', r.status_code, ' / '.join(
                    [*(f[0] for f in folder_path), title]), attrs['href'], new_location)
                attrs['href'] = new_location
                changed.append(new_data)
            elif r.status_code == HTTPStatus.NOT_FOUND:
                log.error('%d: "%s" @ "%s"', r.status_code,
                          ' / '.join([*(f[0] for f in folder_path), title]), attrs['href'])
                not_found.append(new_data)
        ref.append(new_data)

    recurse_bookmarks_html(Soup(html_content, 'html5lib'), callback)
    return data, changed, not_found



"""Typing helpers."""
from __future__ import annotations

from enum import IntEnum
from os import PathLike
from typing import TYPE_CHECKING, Annotated, Any, Literal, NotRequired, TypeVar, TypedDict
import os
import typing

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = ('CDStatus', 'DecodeErrorsOption', 'FileDescriptorOrPath', 'INCITS38Code', 'ProbeDict',
           'StrOrBytesPath', 'StrPath', 'StreamDispositionDict', 'StreamsDict', 'UNIXStrPath',
           'assert_not_none', 'contains_type_path_like_str')

DecodeErrorsOption = Literal['ignore', 'replace', 'strict']
"""Decode errors option for :py:func:`deltona.utils.decode`."""
INCITS38Code = Literal['AK', 'AL', 'AR', 'AS', 'AZ', 'CA', 'CO', 'CT', 'DC', 'DE', 'FL', 'FM', 'GA',
                       'GU', 'HI', 'IA', 'ID', 'IL', 'IN', 'KS', 'KY', 'LA', 'MA', 'MD', 'ME', 'MH',
                       'MI', 'MN', 'MO', 'MP', 'MS', 'MT', 'NC', 'ND', 'NE', 'NH', 'NJ', 'NM', 'NV',
                       'NY', 'OH', 'OK', 'OR', 'PA', 'PR', 'PW', 'RI', 'SC', 'SD', 'TN', 'TX', 'UM',
                       'UT', 'VA', 'VI', 'VT', 'WA', 'WI', 'WV', 'WY']
"""Two-letter state code according to INCITS 38-2009."""
StrOrBytesPath = str | bytes | PathLike[str] | PathLike[bytes]
"""String, bytes, ``PathLike[str]`` or ``PathLike[bytes]``."""
StrPath = str | PathLike[str]
"""String or ``PathLike[str]``."""
FileDescriptorOrPath = int | StrOrBytesPath
"""File descriptor or path."""
UNIXStrPath = Annotated[StrPath, 'unix']
"""String or ``PathLike[str]`` that is a UNIX path."""
StrPathMustExist = Annotated[StrPath, 'must_exist']
"""String or ``PathLike[str]`` that must exist."""
_T = TypeVar('_T')


class CDStatus(IntEnum):
    """CD status codes."""
    DISC_OK = 4
    """Disc is OK."""
    DRIVE_NOT_READY = 3
    """Drive is not ready."""
    NO_DISC = 1
    """No disc in drive."""
    NO_INFO = 0
    """No information available."""
    TRAY_OPEN = 2
    """Tray is open."""


def contains_type_path_like_str(type_hints: Any) -> bool:  # pragma: no cover
    """Check if a type hint contains ``os.PathLike[str]``."""
    return os.PathLike[str] in typing.get_args(type_hints)


def assert_not_none(var: _T | None) -> _T:  # pragma: no cover
    """
    Assert the ``var`` is not None and return it.

    This will remove ``| None`` from type ``_T``.
    """
    assert var is not None
    return var


# Used by chrome-bisect-flags
class ChromeLocalStateBrowser(TypedDict):
    """Chrome ``local state[browser]`` dictionary."""
    enabled_labs_experiments: Sequence[str]
    """Enabled labs experiments, also known as flags"""


class ChromeLocalState(TypedDict):
    """Chrome local state dictionary."""
    browser: ChromeLocalStateBrowser
    """Browser."""


class StreamDispositionDict(TypedDict):
    """FFmpeg stream disposition dictionary."""
    default: Literal[0, 1]
    """Default stream."""


class TagsDict(TypedDict):
    """Tags dictionary."""
    info_json: NotRequired[str]
    """Info JSON."""
    TXXX: NotRequired[str]
    """TXXX tag."""


class StreamsDict(TypedDict):
    """FFmpeg stream dictionary."""
    codec_type: Literal['audio', 'video']
    """Codec type."""
    disposition: StreamDispositionDict
    """Stream disposition dictionary."""
    height: int
    """Height of the video stream."""
    tags: TagsDict
    """Tags dictionary."""
    width: int
    """Width of the video stream."""


class FormatDict(TypedDict):
    """FFmpeg format dictionary."""
    tags: TagsDict
    """Tags dictionary."""


class ProbeDict(TypedDict):
    """FFmpeg probe result returned by :py:func:`deltona.media.ffprobe`."""
    format: FormatDict
    """Format dictionary."""
    streams: Sequence[StreamsDict]
    """Dictionary of streams."""



"""Salary calculator."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import UTC, datetime
from typing import TYPE_CHECKING, TypedDict, cast, override

import requests

from .string import strip_ansi_if_no_colors

if TYPE_CHECKING:
    from .typing import INCITS38Code

__all__ = ('SalaryResponse', 'calculate_salary')

# Find the API key by looking at requests on this page
# http://www.symmetry.com/try-it-for-free/calculators
API_KEY = 'RnFqNFA0NVlRTExEenRwWjNiRnJrTXY4WkZHZEpkcENEeFFzQ3F0Nnh5VT0='
POST_URI = ('https://calculators.symmetry.com/api/calculators/'
            'hourly?report=none')
REFERER = 'https://www.symmetry.com/'


class ContentDict(TypedDict):
    federal: float
    fica: float
    medicare: float
    netPay: float
    state: float


class ResponseDict(TypedDict):
    content: ContentDict


@dataclass
class SalaryResponse:
    """Response from the Symmetry API."""
    federal: float
    """Federal tax amount."""
    fica: float
    """FICA tax amount."""
    fuckery: float
    """Difference between gross and net pay."""
    gross: float
    """Gross pay amount."""
    medicare: float
    """Medicare tax amount."""
    net_pay: float
    """Net pay amount."""
    state: float
    """State tax amount."""
    @override
    def __str__(self) -> str:
        return strip_ansi_if_no_colors(f"""Gross     \033[1;32m{self.gross:8.2f}\033[0m
Federal   \033[1;32m{self.federal:8.2f}\033[0m
FICA      \033[1;32m{self.fica:8.2f}\033[0m
Medicare  \033[1;32m{self.medicare:8.2f}\033[0m
State     \033[1;32m{self.state:8.2f}\033[0m
------------------
Net       \033[1;32m{self.net_pay:8.2f}\033[0m

------------------
Fuckery   \033[1;31m{self.fuckery:8.2f}\033[0m""")


def calculate_salary(*,
                     hours: int = 70,
                     pay_rate: float = 70.0,
                     state: INCITS38Code = 'FL') -> SalaryResponse:
    """
    Calculate a US salary using the Symmetry API.

    Parameters
    ----------
    hours : int
        The number of hours worked. Default is 70.
    pay_rate : float
        The pay rate per hour. Default is 70.0.
    state : INCITS38Code
        The state code. Default is 'FL'.

    Returns
    -------
    SalaryResponse
        The response from the Symmetry API.
    """
    check_date = int(datetime.now(tz=UTC).timestamp() * 1000)
    gross_pay = hours * pay_rate
    req = requests.post(POST_URI,
                        headers={
                            'accept': 'application/json, text/javascript, */*; q=0.01',
                            'cache-control': 'no-cache',
                            'dnt': '1',
                            'origin': 'https://www.symmetry.com',
                            'pcc-api-key': API_KEY,
                            'referer': REFERER
                        },
                        json={
                            'checkDate': check_date,
                            'state': state.upper(),
                            'rates': [{
                                'payRate': str(pay_rate),
                                'hours': str(hours),
                            }],
                            'grossPay': str(gross_pay),
                            'grossPayType': 'PAY_PER_PERIOD',
                            'grossPayYTD': '0',
                            'payFrequency': 'MONTHLY',
                            'exemptFederal': 'false',
                            'exemptFica': 'false',
                            'exemptMedicare': 'false',
                            'federalFilingStatusType': 'SINGLE',
                            'federalAllowances': '0',
                            'additionalFederalWithholding': '0',
                            'roundFederalWithholding': 'false',
                            'print': {
                                'checkDate': check_date,
                                'checkNumber': '',
                                'checkNumberOnCheck': 'false',
                                'companyAddressLine1': '',
                                'companyAddressLine2': '',
                                'companyAddressLine3': '',
                                'companyName': '',
                                'companyNameOnCheck': 'false',
                                'employeeAddressLine1': '',
                                'employeeAddressLine2': '',
                                'employeeAddressLine3': '',
                                'employeeName': '',
                                'id': '',
                                'remarks': ''
                            },
                            'otherIncome': [],
                            'payCodes': [],
                            'stockOptions': [],
                            'stateInfo': {
                                'parms': [{
                                    'name': 'TOTALALLOWANCES',
                                    'value': '0'
                                }, {
                                    'name': 'additionalStateWithholding',
                                    'value': '0'
                                }, {
                                    'name': 'SPOUSEBLINDNESS',
                                    'value': 'false'
                                }, {
                                    'name': 'stateExemption',
                                    'value': 'false'
                                }, {
                                    'name': 'PERSONALBLINDNESS',
                                    'value': 'false'
                                }, {
                                    'name': 'HEADOFHOUSEHOLD',
                                    'value': 'false'
                                }, {
                                    'name': 'FULLTIMESTUDENT',
                                    'value': 'false'
                                }]
                            },
                            'voluntaryDeductions': [],
                            'presetDeductions': []
                        },
                        timeout=30)
    req.raise_for_status()
    data = cast('ResponseDict', req.json())['content']
    return SalaryResponse(federal=data['federal'],
                          fica=data['fica'],
                          fuckery=gross_pay - data['netPay'],
                          gross=gross_pay,
                          medicare=data['medicare'],
                          net_pay=data['netPay'],
                          state=data['state'])



"""Uncategorised utilities."""
# ruff: noqa: N815
from __future__ import annotations

from http import HTTPStatus
from io import BytesIO
from math import trunc
from os import environ
from pathlib import Path
from shlex import quote
from shutil import copyfile, rmtree, which
from signal import SIGTERM
from typing import TYPE_CHECKING, Literal, NamedTuple, overload, override
import csv
import logging
import os
import re
import sqlite3
import struct
import subprocess as sp
import tarfile
import tempfile
import time

from requests.adapters import BaseAdapter
import platformdirs
import requests
import xz

from .media import CD_FRAMES
from .system import IS_WINDOWS, kill_wine
from .windows import (
    LF_FULLFACESIZE,
    CharacterSet,
    ClipPrecision,
    Family,
    OutputPrecision,
    Pitch,
    Quality,
    Weight,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from paramiko import SFTPClient, SSHClient

    from .typing import StrPath

__all__ = ('WineWindowsVersion', 'add_cdda_times', 'create_wine_prefix', 'secure_move_path')

ZERO_TO_99 = '|'.join(f'{x:02d}' for x in range(100))
ZERO_TO_59 = '|'.join(f'{x:02d}' for x in range(60))
ZERO_TO_74 = '|'.join(f'{x:02d}' for x in range(75))
TIMES_RE = re.compile(f'^({ZERO_TO_99}):({ZERO_TO_59}):({ZERO_TO_74})$')
MAX_MINUTES = 99
MAX_SECONDS = 60
log = logging.getLogger(__name__)


def add_cdda_times(times: Iterable[str] | None) -> str | None:
    """
    Add CDDA time strings and get a total runtime in CDDA format.

    CDDA format is ``MM:SS:FF`` where ``MM`` is minutes, ``SS`` is seconds, and ``FF`` is frames.
    """
    if not times:
        return None
    total_ms = 0
    for time_ in times:
        if not (res := re.match(TIMES_RE, time_)):
            return None
        minutes, seconds, frames = [int(x) for x in res.groups()]
        total_ms += (minutes * MAX_SECONDS * 1000) + (seconds * 1000) + (
            (frames * 1000) // CD_FRAMES)
    minutes = total_ms // (MAX_SECONDS * 1000)
    remainder_ms = total_ms % (MAX_SECONDS * 1000)
    seconds = remainder_ms // 1000
    remainder_ms %= 1000
    frames = round((remainder_ms * 1000 * CD_FRAMES) / 1000000)
    if minutes > MAX_MINUTES or seconds > (MAX_SECONDS - 1) or frames > (CD_FRAMES - 1):
        return None
    return f'{trunc(minutes):02d}:{trunc(seconds):02d}:{trunc(frames):02d}'


WINETRICKS_VERSION_MAPPING = {
    '11': 'win11',
    '10': 'win10',
    'vista': 'vista',
    '2k3': 'win2k3',
    '7': 'win7',
    '8': 'win8',
    'xp': 'winxp',
    '81': 'win81',
    # 32-bit only
    '2k': 'win2k',
    '98': 'win98',
    '95': 'win95'
}
"""Mapping of Windows versions to winetricks versions."""
DEFAULT_DPI = 96
"""Default DPI for Wine prefixes."""
_CREATE_WINE_PREFIX_NOTO_FONT_REPLACEMENTS = {
    'Arial Baltic,186', 'Arial CE,238', 'Arial CYR,204', 'Arial Greek,161', 'Arial TUR,162',
    'Courier New Baltic,186', 'Courier New CE,238', 'Courier New CYR,204', 'Courier New Greek,161',
    'Courier New TUR,162', 'Helv', 'Helvetica', 'MS Shell Dlg', 'MS Shell Dlg 2', 'MS Sans Serif',
    'Segoe UI', 'System', 'Tahoma', 'Times', 'Times New Roman Baltic,186', 'Times New Roman CE,238',
    'Times New Roman CYR,204', 'Times New Roman Greek,161', 'Times New Roman TUR,162', 'Tms Rmn',
    'Verdana'
}
_CREATE_WINE_PREFIX_NOTO_REGISTRY_ENTRIES = {
    'Caption', 'Icon', 'Menu', 'Message', 'SmCaption', 'Status'
}
WineWindowsVersion = Literal['11', '10', 'vista', '2k3', '7', '8', 'xp', '81', '2k', '98', '95']
"""Windows versions supported by Wine."""


class LOGFONTW(NamedTuple):
    """Windows LOGFONTW structure as a named tuple."""
    lfHeight: int
    lfWidth: int
    lfEscapement: int
    lfOrientation: int
    lfWeight: int
    lfItalic: bool
    lfUnderline: bool
    lfStrikeOut: bool
    lfCharSet: int
    lfOutPrecision: int
    lfClipPrecision: int
    lfQuality: int
    lfPitchAndFamily: int


Q4WINE_DEFAULT_ICONS: tuple[tuple[str, str, str, str, str, str], ...] = (
    ('', 'control.exe', 'control', 'Wine control panel', 'system', 'Control Panel'),
    ('', 'eject.exe', 'eject', 'Wine CD eject tool', 'system', 'Eject'),
    ('', 'explorer.exe', 'explorer', 'Browse the files in the virtual Wine Drive', 'system',
     'Explorer'),
    ('', 'iexplore.exe', 'iexplore', 'Wine internet browser', 'system', 'Internet Explorer'),
    ('', 'notepad.exe', 'notepad', 'Wine notepad text editor', 'system', 'Notepad'),
    ('', 'oleview.exe', 'oleview', 'Wine OLE/COM object viewer', 'system', 'OLE Viewer'),
    ('', 'regedit.exe', 'regedit', 'Wine registry editor', 'system', 'Registry Editor'),
    ('', 'taskmgr.exe', 'taskmgr', 'Wine task manager', 'system', 'Task Manager'),
    ('', 'uninstaller.exe', 'uninstaller', 'Uninstall Windows programs under Wine properly',
     'system', 'Uninstaller'),
    ('', 'winecfg.exe', 'winecfg', 'Configure the general settings for Wine', 'system',
     'Configuration'),
    ('', 'wineconsole', 'wineconsole', 'Wineconsole is similar to wine command wcmd', 'system',
     'Console'),
    ('', 'winemine.exe', 'winemine', 'Wine sweeper game', 'system', 'Winemine'),
    ('', 'wordpad.exe', 'wordpad', 'Wine wordpad text editor', 'system', 'WordPad'),
)
"""Shortcuts to add to Q4Wine for the prefix."""


def create_wine_prefix(prefix_name: str,
                       *,
                       _32bit: bool = False,
                       asio: bool = False,
                       disable_explorer: bool = False,
                       disable_services: bool = False,
                       dpi: int = DEFAULT_DPI,
                       dxva_vaapi: bool = False,
                       dxvk_nvapi: bool = False,
                       eax: bool = False,
                       gtk: bool = False,
                       no_associations: bool = False,
                       no_gecko: bool = False,
                       no_mono: bool = False,
                       no_xdg: bool = False,
                       noto_sans: bool = False,
                       prefix_root: StrPath | None = None,
                       sandbox: bool = False,
                       tmpfs: bool = False,
                       tricks: Iterable[str] | None = None,
                       vd: str = 'off',
                       windows_version: WineWindowsVersion = '10',
                       winrt_dark: bool = False) -> Path:
    """
    Create a Wine prefix with custom settings.

    If winetricks is not installed, the ``tricks`` argument will be ignored.

    Parameters
    ----------
    prefix_name : str
        Name of the prefix to create.
    _32bit : bool
        Create a 32-bit Wine prefix.
    asio : bool
        Enable ASIO support.
    disable_explorer : bool
        Disable ``explorer.exe`` from automatically starting.
    disable_services : bool
        Disable ``services.exe`` from automatically starting.
    dpi : int
        Screen DPI.
    dxva_vaapi : bool
        Enable VAAPI support for DXVA.
    dxvk_nvapi : bool
        Enable DXVK NVAPI support.
    eax : bool
        Enable EAX support.
    gtk : bool
        Enable GTK theme support.
    no_associations : bool
        Disable file associations.
    no_gecko : bool
        Disable Gecko support.
    no_mono : bool
        Disable Mono support.
    no_xdg : bool
        Disable XDG support.
    noto_sans : bool
        Use Noto Sans font.
    prefix_root : StrPath | None
        Root directory for the prefix. If ``None``, defaults to ``~/.local/share/wineprefixes``.
    sandbox : bool
        Enable sandbox mode.
    tmpfs : bool
        Use ``/tmp`` as a temporary filesystem.
    tricks : Iterable[str] | None
        List of winetricks to run. If ``None``, defaults to an empty list.
    vd : str
        Virtual desktop mode. If ``'off'``, disables virtual desktop mode.
    windows_version : WineWindowsVersion
        Windows version to set for the prefix.
    winrt_dark : bool
        Enable Windows 10 dark mode.

    Returns
    -------
    Path
        Path to the created prefix.

    Raises
    ------
    FileExistsError
    """
    tricks = list((t for t in tricks
                   if t not in WINETRICKS_VERSION_MAPPING.values() and not t.startswith('vd=')
                   ) if tricks else [])
    prefix_root = Path(prefix_root) if prefix_root else Path.home() / '.local/share/wineprefixes'
    prefix_root.mkdir(parents=True, exist_ok=True)
    target = prefix_root / prefix_name
    if target.exists():
        raise FileExistsError
    arch = 'win32' if _32bit else None
    if 'DISPLAY' not in environ or 'XAUTHORITY' not in environ:
        log.warning('Wine will likely fail to run since DISPLAY or XAUTHORITY are not in the '
                    'environment.')
    esync = environ.get('WINEESYNC', '')
    env = {
        'DISPLAY': environ.get('DISPLAY', ''),
        'PATH': environ['PATH'],
        'WINEPREFIX': str(target),
        'XAUTHORITY': environ.get('XAUTHORITY', ''),
        'WINEDEBUG': environ.get('WINEDEBUG', 'fixme-all')
    } | ({
        'WINEARCH': environ.get('WINEARCH', arch)
    } if arch else {}) | ({
        'WINEESYNC': esync
    } if esync else {})
    # Warm up Wine
    cmd: tuple[str, ...] = ('wineboot', '--init')
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, env=env, check=True)
    # Wait for the server to finish because wineboot does not necessarily quit before the prefix is
    # done being set up.
    cmd = ('wineserver', '-w')
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, env=env, check=True)
    if dpi != DEFAULT_DPI:
        cmd = ('wine', 'reg', 'add', r'HKCU\Control Panel\Desktop', '/t', 'REG_DWORD', '/v',
               'LogPixels', '/d', str(dpi), '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if dxva_vaapi:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DXVA2', '/v', 'backend', '/d', 'va', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if eax:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DirectSound', '/v', 'EAXEnabled', '/d',
               'Y', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if gtk:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine', '/v', 'ThemeEngine', '/d', 'GTK', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if winrt_dark:
        for k in ('AppsUseLightTheme', 'SystemUsesLightTheme'):
            cmd = ('wine', 'reg', 'add',
                   r'HKCU\Software\Microsoft\Windows\CurrentVersion\Themes\Personalize', '/t',
                   'REG_DWORD', '/v', k, '/d', '0', '/f')
            log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, env=env, check=True)
    if no_associations:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\Explorer\FileAssociations', '/v',
               'Enable', '/d', 'N', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if no_xdg:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v',
               'winemenubuilder.exe', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if no_mono:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', 'mscoree', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if no_gecko:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', 'mshtml', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if disable_explorer:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', 'explorer.exe', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if disable_services:
        cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', 'services.exe', '/f')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
    if tmpfs:
        username = environ.get('USER', environ.get('USERNAME', 'user'))
        rmtree(target / f'drive_c/users/{username}/Temp', ignore_errors=True)
        rmtree(target / 'drive_c/windows/temp', ignore_errors=True)
        Path(target / f'drive_c/users/{username}/Temp').symlink_to(tempfile.gettempdir(),
                                                                   target_is_directory=True)
        Path(target / 'drive_c/windows/temp').symlink_to(tempfile.gettempdir(),
                                                         target_is_directory=True)
    if dxvk_nvapi:
        tricks += ['dxvk']
    try:
        tricks += [WINETRICKS_VERSION_MAPPING[windows_version]]
        if sandbox:
            tricks += ['isolate_home', 'sandbox']
        if vd != 'off':
            tricks += [f'vd={vd}']
        if (winetricks := which('winetricks')):
            cmd = (winetricks, '--force', '--country=US', '--unattended', f'prefix={prefix_name}',
                   *sorted(set(tricks)))
            log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, check=True)
    except sp.CalledProcessError as e:  # pragma: no cover
        log.warning('Winetricks exit code was %d but it may have succeeded.', e.returncode)
    if dxvk_nvapi:
        cmd = ('setup_vkd3d_proton.sh', 'install')
        log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
        sp.run(cmd, env=env, check=True)
        version = '0.8.3'
        nvidia_libs = 'nvidia-libs'
        prefix = f'{nvidia_libs}-{version}'
        r = requests.get(
            f'https://github.com/SveSop/{nvidia_libs}/releases/download/v{version}/{prefix}.tar.xz',
            timeout=15)
        r.raise_for_status()
        with xz.open(BytesIO(r.content)) as xz_file, tarfile.TarFile(fileobj=xz_file) as tar:
            for item in ('nvcuda', 'nvcuvid', 'nvencodeapi', 'nvapi'):
                cmd = ('wine', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', item, '/d',
                       'native', '/f')
                log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
                sp.run(cmd, env=env, check=True)
                member = tar.getmember(f'{prefix}/x32/{item}.dll')
                member.name = f'{item}.dll'
                tar.extract(member, target / 'drive_c' / 'windows' / 'syswow64')
            if not _32bit:
                for item in ('nvcuda', 'nvoptix', 'nvcuvid', 'nvencodeapi64', 'nvapi64',
                             'nvofapi64'):
                    cmd = ('wine64', 'reg', 'add', r'HKCU\Software\Wine\DllOverrides', '/v', item,
                           '/d', 'native', '/f')
                    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
                    sp.run(cmd, env=env, check=True)
                    member = tar.getmember(f'{prefix}/x64/{item}.dll')
                    member.name = f'{item}.dll'
                    tar.extract(member, target / 'drive_c' / 'windows' / 'system32')
        for prefix in ('', '_'):
            copyfile(f'/lib64/nvidia/wine/{prefix}nvngx.dll',
                     target / 'drive_c' / 'windows' / 'system32' / f'{prefix}nvngx.dll')
        if not _32bit:
            cmd = ('wine64', 'reg', 'add', r'HKLM\Software\NVIDIA Corporation\Global\NGXCore', '/t',
                   'REG_SZ', '/v', 'FullPath', '/d', r'C:\Windows\system32', '/f')
            log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, env=env, check=True)
    if noto_sans:
        for font_name in _CREATE_WINE_PREFIX_NOTO_FONT_REPLACEMENTS:
            cmd = ('wine', 'reg', 'add',
                   r'HKLM\Software\Microsoft\Windows NT\CurrentVersion\FontSubstitutes', '/t',
                   'REG_SZ', '/v', font_name, '/d', 'Noto Sans', '/f')
            log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, env=env, check=True)
        face_name = 'Noto Sans'.encode('utf-16le').ljust(LF_FULLFACESIZE, b'\0')
        for entry_name in _CREATE_WINE_PREFIX_NOTO_REGISTRY_ENTRIES:
            cmd = (
                'wine',
                'reg',
                'add',
                r'HKCU\Control Panel\Desktop\WindowMetrics',
                '/t',
                'REG_BINARY',
                '/v',
                f'{entry_name}Font',
                '/d',
                ''.join(f'{x:02x}' for x in struct.pack(
                    '=5l8B64B',
                    *LOGFONTW(
                        lfHeight=-12,  # Size 9 pt
                        lfWidth=0,
                        lfEscapement=0,
                        lfOrientation=0,
                        lfWeight=Weight.FW_BOLD if entry_name == 'Caption' else Weight.FW_NORMAL,
                        lfItalic=False,
                        lfUnderline=False,
                        lfStrikeOut=False,
                        lfCharSet=CharacterSet.DEFAULT_CHARSET,
                        lfOutPrecision=OutputPrecision.OUT_DEFAULT_PRECIS,
                        lfClipPrecision=ClipPrecision.CLIP_DEFAULT_PRECIS,
                        lfQuality=Quality.DEFAULT_QUALITY,
                        lfPitchAndFamily=Pitch.VARIABLE_PITCH | Family.FF_SWISS),
                    *face_name)),
                '/f')
            log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
            sp.run(cmd, env=env, check=True)
    if asio:
        if register := which('wineasio-register'):
            log.debug('Running: %s', register)
            sp.run((register,), env=env, check=True)
        else:
            log.warning('Skipping ASIO setup because wineasio-register is not in PATH.')
    if (db_path := (platformdirs.user_config_path() / 'q4wine/db/generic.dat')).exists():
        # Based on addPrefix() and createPrefixDBStructure().
        # https://github.com/brezerk/q4wine/blob/master/src/core/database/prefix.cpp#L250
        # https://github.com/brezerk/q4wine/blob/master/src/q4wine-lib/q4wine-lib.cpp#L1920
        log.debug('Adding this prefix to Q4Wine.')
        with sqlite3.connect(db_path) as conn:
            c = conn.cursor()
            c.execute(
                'INSERT INTO prefix (name, path, mountpoint_windrive, run_string, version_id) '
                'VALUES (?, ?, ?, ?, 1)',
                (prefix_name, str(target), 'D:',
                 r'%CONSOLE_BIN% %CONSOLE_ARGS% %ENV_BIN% %ENV_ARGS% /bin/sh -c '
                 r'"%WORK_DIR% %SET_NICE% %WINE_BIN% %VIRTUAL_DESKTOP% %PROGRAM_BIN% '
                 r'%PROGRAM_ARGS% 2>&1 "'))
            prefix_id = c.lastrowid
            log.debug('Q4Wine prefix ID: %d', prefix_id)
            assert prefix_id is not None
            for dir_name in ('system', 'autostart', 'import'):
                c.execute('INSERT INTO dir (name, prefix_id) VALUES (?, ?)', (dir_name, prefix_id))
            for args, exec_, icon_path, desc, folder, display_name in Q4WINE_DEFAULT_ICONS:
                c.execute(
                    """INSERT INTO icon (
    cmdargs, exec, icon_path, desc, dir_id, name, prefix_id, nice)
    VALUES (
        ?, ?, ?, ?, (
            SELECT id FROM dir WHERE name = ? AND prefix_id = ?
        ), ?, ?, 0
    )""", (args or None, exec_, icon_path, desc, folder, prefix_id, display_name, prefix_id))
            c.execute('DELETE FROM logging WHERE prefix_id = ?', (prefix_id,))
    return target


def unregister_wine_file_associations(*, debug: bool = False) -> None:
    """Unregister all Wine file associations."""
    kill_wine()
    for item in (Path.home() / '.local/share/applications').glob('wine-extension-*.desktop'):
        log.debug('Removing file association "%s".', item)
        item.unlink()
    for item in (Path.home() / '.local/share/icons/hicolor').rglob('application-x-wine-extension*'):
        log.debug('Removing icon "%s".', item)
        item.unlink()
    (Path.home() / '.local/share/applications/mimeinfo.cache').unlink(missing_ok=True)
    for item in (Path.home() / '.local/share/mime/packages').glob('x-wine*'):
        log.debug('Removing MIME file "%s".', item)
        item.unlink()
    for item in (Path.home() / '.local/share/application').glob('x-wine-extension*'):
        log.debug('Removing MIME file "%s".', item)
        item.unlink()
    cmd: tuple[str, ...] = ('update-desktop-database', *(('-v',) if debug else ()),
                            str(Path.home() / '.local/share/applications'))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    cmd = ('update-mime-database', *(('-v',) if debug else
                                     ()), str(Path.home() / '.local/share/mime'))
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)


def secure_move_path(client: SSHClient,
                     filename: StrPath,
                     remote_target: str,
                     *,
                     dry_run: bool = False,
                     preserve_stats: bool = False,
                     write_into: bool = False) -> None:
    """Like ``scp`` but moves the file."""
    log.debug('Source: "%s", remote target: "%s"', filename, remote_target)

    def mkdir_ignore_existing(sftp: SFTPClient, td: str, times: tuple[float, float]) -> None:
        if not write_into:
            log.debug('MKDIR "%s"', td)
            if not dry_run:
                sftp.mkdir(td)
                if preserve_stats:
                    sftp.utime(td, times)
            return
        try:
            sftp.stat(td)
        except FileNotFoundError:
            log.debug('MKDIR "%s"', td)
            if not dry_run:
                sftp.mkdir(td)
                if preserve_stats:
                    sftp.utime(td, times)

    path = Path(filename)
    _, stdout, __ = client.exec_command('echo "${HOME}"')
    remote_target = remote_target.replace('~', stdout.read().decode().strip())
    with client.open_sftp() as sftp:
        if path.is_file():
            if not dry_run:
                sftp.put(filename, remote_target)
                if preserve_stats:
                    local_s = Path(filename).stat()
                    sftp.utime(remote_target, (local_s.st_atime, local_s.st_mtime))
            log.debug('Deleting local file "%s".', path)
            if not dry_run:
                path.unlink()
        else:
            pf = Path(filename)
            pf_stat = pf.stat()
            bn_filename = pf.name
            dn_prefix = str(pf).replace(bn_filename, '')
            mkdir_ignore_existing(sftp, remote_target, (pf_stat.st_atime, pf_stat.st_mtime))
            for root, dirs, files in os.walk(filename, followlinks=True):
                p_root = Path(root)
                remote_target_dir = f'{remote_target}/{bn_filename}'
                p_root_stat = p_root.stat()
                mkdir_ignore_existing(sftp, remote_target_dir,
                                      (p_root_stat.st_atime, p_root_stat.st_mtime))
                for name in sorted(dirs):
                    p_root_stat = (p_root / name).stat()
                    dp = str(p_root / name).replace(dn_prefix, '')
                    remote_target_dir = f'{remote_target}/{dp}'
                    mkdir_ignore_existing(sftp, remote_target_dir,
                                          (p_root_stat.st_atime, p_root_stat.st_mtime))
                for name in sorted(files):
                    src = p_root / name
                    dp = str(p_root / name).replace(dn_prefix, '')
                    log.debug('PUT "%s" "%s/%s"', src, remote_target, dp)
                    if not dry_run:
                        sftp.put(src, f'{remote_target}/{dp}')
                        if preserve_stats:
                            local_s = Path(src).stat()
                            sftp.utime(f'{remote_target}/{dp}',
                                       (local_s.st_atime, local_s.st_mtime))
            if not dry_run:
                rmtree(filename, ignore_errors=True)
            else:
                log.debug('Would delete local directory "%s".', filename)


@overload
def kill_processes_by_name(name: str) -> None:  # pragma: no cover
    pass


@overload
def kill_processes_by_name(name: str,
                           wait_timeout: float,
                           signal: int = SIGTERM,
                           *,
                           force: bool = False) -> list[int]:  # pragma: no cover
    pass


def kill_processes_by_name(name: str,
                           wait_timeout: float | None = None,
                           signal: int = SIGTERM,
                           *,
                           force: bool = False) -> list[int] | None:
    """
    Terminate processes by name.

    Alternative to using `psutil <https://pypi.org/project/psutil/>`_.

    Parameters
    ----------
    name : str
        Process name (base name) or image name (Windows).
    wait_timeout : float | None
        If set and processes remain after ending processes, wait this amount of time in seconds.
    signal : int
        Signal to use. Only applies to non-Windows.
    force : bool
        If ``wait_timeout`` is set and ``True``, forcefully end the processes after the wait time.

    Returns
    -------
    list[int] | None
        PIDs of processes that may still be running, or ``None`` if ``wait_timeout`` is not
        specified.
    """
    name = f'{name}{Path(name).suffix or ".exe"}' if IS_WINDOWS else name
    pids: list[int] = []
    if IS_WINDOWS:
        sp.run(('taskkill.exe', '/im', name), check=False, capture_output=True)
    else:
        sp.run(('killall', f'-{signal}', name), check=False, capture_output=True)
    if wait_timeout:
        lines = sp.run(
            ('tasklist.exe', '/fo', 'csv', '/fi', f'IMAGENAME eq {name}') if IS_WINDOWS else
            ('ps', 'ax'),
            check=True,
            capture_output=True,
            text=True).stdout.splitlines()
        if pids := [int(x[1]) for x in list(csv.reader(lines))[1:]] if IS_WINDOWS else [
                int(y[0]) for y in (x.split() for x in lines) if Path(y[0]).name == name
        ]:
            time.sleep(wait_timeout)
            if force:
                sp.run(('taskkill.exe', *(t for sl in (('/pid', str(pid)) for pid in pids)
                                          for t in sl), '/f') if IS_WINDOWS else
                       ('kill', '-9', *(str(x) for x in pids)),
                       check=False,
                       capture_output=True)
    return pids if wait_timeout else None


class DataAdapter(BaseAdapter):
    """
    Adapter for requests to handle data: URLs.

    Example use:

    .. code-block:: python
       s = requests.Session()
       s.mount('data:', DataAdapter())
    """
    @override
    def send(self,
             request: requests.PreparedRequest,
             stream: bool = False,
             timeout: float | tuple[float, float] | tuple[float, None] | None = None,
             verify: bool | str = True,
             cert: bytes | str | tuple[bytes | str, bytes | str] | None = None,
             proxies: Mapping[str, str] | None = None) -> requests.Response:
        r = requests.Response()
        assert request.url is not None
        r._content = request.url[5:].encode()  # noqa: SLF001
        r.status_code = HTTPStatus.OK
        return r

    @override
    def close(self) -> None:
        pass



"""String utilities."""
# ruff: noqa: RUF001
from __future__ import annotations

from functools import cache
from itertools import batched, takewhile
from pathlib import Path
from typing import TYPE_CHECKING, cast
import os
import re
import string

from .typing import StrPath, assert_not_none

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence

__all__ = ('fullwidth_to_narrow', 'hexstr2bytes', 'hexstr2bytes_generator', 'is_ascii', 'is_url',
           'sanitize', 'slugify', 'strip_ansi', 'strip_ansi_if_no_colors', 'underscorize',
           'unix_path_to_wine')

ORD_MAX = 128
STRIP_ANSI_PATTERN = re.compile(r'\x1B\[\d+(;\d+){0,2}m')


@cache
def strip_ansi(o: str) -> str:
    """
    Remove ANSI escape sequences from `o`.

    As defined by ECMA-048 in http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-048.

    Taken from https://github.com/ewen-lbh/python-strip-ansi/ due to installation issues with
    Poetry.
    """
    return STRIP_ANSI_PATTERN.sub('', o)


def strip_ansi_if_no_colors(s: str) -> str:
    """
    Strip ANSI colour-codes if the ``NO_COLOR`` environment variable is set.

    See https://no-color.org/.
    """
    return strip_ansi(s) if os.environ.get('NO_COLOR') else s


@cache
def underscorize(s: str) -> str:
    """Replace all space-type characters with ``_``."""
    return re.sub(r'\s+', '_', s)


@cache
def is_ascii(s: Sequence[str]) -> bool:
    """Check if a string consists of only ASCII characters."""
    return len(s) == len(list(takewhile(lambda x: ord(x) < ORD_MAX, s)))


def hexstr2bytes_generator(s: str) -> Iterator[int]:
    """
    Convert a hex string such as ``"01020a"`` to integers.

    Yields
    ------
    int
        The integers in the hex string.

    Raises
    ------
    ValueError
        If the string is not a valid hex string.
    """
    for hex_num in batched(s, 2):
        if len(hex_num) != 2:  # noqa: PLR2004
            raise ValueError(hex_num)
        yield int(''.join(hex_num), 16)


def hexstr2bytes(s: str) -> bytes:
    """Convert a hex string such as ``"01020a"`` to its bytes form (``0x1 0x2 0x10)``)."""
    return bytes(hexstr2bytes_generator(s))


@cache
def unix_path_to_wine(path: StrPath) -> str:
    """
    Convert a UNIX path to an absolute Wine path.

    If the path does not exist, the output will be the current path and the path passed in combined.

    The output path will begin at letter ``Z:``. Other drive letters are not supported.

    Parameters
    ----------
    path : StrPath
        Path to convert.

    Returns
    -------
    str
        Window-style Wine absolute path.
    """
    try:
        path = Path(path).resolve(strict=True)
    except FileNotFoundError:
        path = Path.cwd() / path
    return f'Z:{path}'.replace('/', '\\')


def _get_yt_dlp_sanitize_filename() -> Callable[..., str]:  # pragma: no cover
    from yt_dlp.utils import sanitize_filename  # noqa: PLC0415
    return sanitize_filename


@cache
def sanitize(s: str, *, restricted: bool = True) -> str:
    """
    Transform a string to a 'sanitised' form.

    Parameters
    ----------
    s : str
        String to transform.
    restricted : bool
        If ``True``, use a restricted form. This is suitable for filenames on Windows.

    Returns
    -------
    str
        Returns a transformed string, which will be at minimum ``'_'``.
    """
    sanitize_filename = _get_yt_dlp_sanitize_filename()
    return re.sub(
        r'([a-z0-9])\-s\-', r'\1s-',
        re.sub(r'\.-', '-',
               re.sub(r'[_\-]+', '-',
                      sanitize_filename(s, restricted=restricted).lower())))


@cache
def is_url(filename: StrPath) -> bool:
    """
    Detect if ``filename`` is a URL.

    This is the same method mpv uses to decide this.
    """
    parts = str(filename).split('://', 1)
    if len(parts) < 2:  # noqa: PLR2004
        return False
    # protocol prefix has no special characters => it's a URL
    return all(x in f'{string.ascii_letters}{string.digits}_' for x in parts[0])


def _get_unidecode_cache_and_unidecode(
) -> tuple[dict[int, Sequence[str | None] | None], Callable[..., str]]:  # pragma: no cover
    from unidecode import Cache, unidecode  # noqa: PLC0415
    return Cache, unidecode


def add_unidecode_custom_replacement(find: str, replace: str) -> None:
    """
    Add a custom replacement to the Unidecode library.

    Call this before calling ``unidecode()``.

    Notes
    -----
    Unidecode is GPL-only which makes anything calling into this significantly also GPL. If you do
    not intend to release GPL code, then you must use a different library such as
    `text-unidecode <https://github.com/kmike/text-unidecode>`_.
    """
    cache, unidecode = _get_unidecode_cache_and_unidecode()
    unidecode(find)  # Force it to load the module
    codepoint = ord(find)
    section = codepoint >> 8
    position = codepoint % 256
    new_section = cast('list[str | None]',
                       (cache[section] if isinstance(cache[section], list) else
                        (list(assert_not_none(cache[section])) if cache[section] is not None else
                         [None for _ in range(position + 1)])))  # convert to mutable type
    assert len(new_section) > position
    new_section[position] = replace
    cache[section] = new_section


FULLWIDTH_MAP = (
    ('　', ' '),
    ('…', '...'),
    ('！', '!'),
    ('？', '?'),
    ('～', '~'),
    ('（', '('),
    ('）', ')'),
    ('＂', '"'),
    ('＃', '#'),
    ('＄', '$'),
    ('％', '%'),
    ('＆', '&'),
    ('＇', "'"),
    ('＊', '*'),
    ('，', ','),
    ('－', '-'),
    ('．', '.'),
    ('／', '/'),
    ('０', '0'),
    ('１', '1'),
    ('２', '2'),
    ('３', '3'),
    ('４', '4'),
    ('５', '5'),
    ('６', '6'),
    ('７', '7'),
    ('８', '8'),
    ('９', '9'),
    ('：', ':'),
    ('；', ';'),
    ('＜', '<'),
    ('＝', '='),
    ('＞', '>'),
    ('？', '?'),
    ('＠', '@'),
    ('Ａ', 'A'),
    ('Ｂ', 'B'),
    ('Ｃ', 'C'),
    ('Ｄ', 'D'),
    ('Ｅ', 'E'),
    ('Ｆ', 'F'),
    ('Ｇ', 'G'),
    ('Ｈ', 'H'),
    ('Ｉ', 'I'),
    ('Ｊ', 'J'),
    ('Ｋ', 'K'),
    ('Ｌ', 'L'),
    ('Ｍ', 'M'),
    ('Ｎ', 'N'),
    ('Ｏ', 'O'),
    ('Ｐ', 'P'),
    ('Ｑ', 'Q'),
    ('Ｒ', 'R'),
    ('Ｓ', 'S'),
    ('Ｔ', 'T'),
    ('Ｕ', 'U'),
    ('Ｖ', 'V'),
    ('Ｗ', 'W'),
    ('Ｘ', 'X'),
    ('Ｙ', 'Y'),
    ('Ｚ', 'Z'),
    ('［', '['),
    ('＼', '\\'),
    ('］', ']'),
    ('＾', '^'),
    ('＿', '_'),
    ('｀', '`'),
    ('ａ', 'a'),
    ('ｂ', 'b'),
    ('ｃ', 'c'),
    ('ｄ', 'd'),
    ('ｅ', 'e'),
    ('ｆ', 'f'),
    ('ｇ', 'g'),
    ('ｈ', 'h'),
    ('ｉ', 'i'),
    ('ｊ', 'j'),
    ('ｋ', 'k'),
    ('ｌ', 'l'),
    ('ｍ', 'm'),
    ('ｎ', 'n'),
    ('ｏ', 'o'),
    ('ｐ', 'p'),
    ('ｑ', 'q'),
    ('ｒ', 'r'),
    ('ｓ', 's'),
    ('ｔ', 't'),
    ('ｕ', 'u'),
    ('ｖ', 'v'),
    ('ｗ', 'w'),
    ('ｘ', 'x'),
    ('ｙ', 'y'),
    ('ｚ', 'z'),
    ('｛', '{'),
    ('｜', '|'),
    ('｝', '}'),
    ('｟', '⸨'),
    ('｠', '⸩'),
    ('￠', '¢'),
    ('￡', '£'),
    ('￢', '¬'),
    ('￣', '‾'),
    ('￤', '|'),
    ('￥', '¥'),
    ('￦', '₩'),
)


@cache
def fullwidth_to_narrow(s: str) -> str:
    """
    Convert fullwidth characters in ``s`` to narrow or halfwidth.

    Unlike Unidecode this will convert ``'￥'`` to its halfwidth form ``'¥'``.
    """
    for find, replace in FULLWIDTH_MAP:
        s = s.replace(find, replace)
    return s


@cache
def slugify(s: str, *, no_lower: bool = False) -> str:
    """Slug string generator."""
    return re.sub(r'[-\s_]+', '-', re.sub(r'[^\w\s-]', '', s if no_lower else s.lower()).strip())


@cache
def is_roman_numeral(string: str) -> bool:
    """Check if a string is a Roman numeral."""
    if not string:
        return False
    return re.match(r'^M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$', string,
                    re.IGNORECASE) is not None


@cache
def fix_apostrophes(word: str) -> str:
    """
    Fix letters around an apostrophe.

    Example: ``"Don'T"`` becomes ``"Don't"``.
    """
    if "'" not in word:
        return word
    return re.sub(r"[A-Za-z]+('[A-Za-z]+)?",
                  lambda mo: f'{mo.group(0)[0].upper()}{mo.group(0)[1:].lower()}', word)


def rev_sentence(w: str) -> str:
    """Reverse a sentence by word."""
    ending = m.group(1) if (m := re.search(r'([\.\!\?])$', w)) else '.'
    lst = list(reversed(re.sub(r'([\.\!\?:;])$', '', w).split()))
    lst[0] = lst[0].title()
    lst[-1] = lst[-1].lower()
    fixed_i = re.sub(r'(\b)i(\b)', r'\1I\2', ' '.join(lst))
    return f'{fixed_i}{ending}'


def rev_sentences(sentences: Sequence[str]) -> Iterator[str]:
    """
    Reverse sentences by word.

    Yields
    ------
    str
        The reversed sentences.
    """
    for line in (x.strip() for x in sentences if x.strip()):
        yield rev_sentence(line)
